/**************************************************************************
   CUTASTRE Accelerated A-contrario Detection of Smooth Trajectories
   Copyright (C) 2014 Remy Abergel (remy.abergel AT gmail.com),
   Lionel Moisan (Lionel.Moisan AT parisdescartes.fr).

   CUTASTRE is an accelerated variant of ASTRE for the detection of
   smooth trajectories in a (noisy) point set sequence. This variant
   manages to break the quadratic time and memory complexity of ASTRE
   with respect to the number of frames, while showing similar (or even
   slightly better) detection performances. The resulting algorithm
   CUTASTRE has a linear complexity with respect to the number of
   frames.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
***************************************************************************/

/*

  Compilation : gcc -O3 cutastre_noholes.c -lm -fopenmp -o cutastre_noholes

  References :

  + [1] ASTRE    : Maël Primet and Lionel Moisan, ``Point tracking: an a-contrario approach'', preprint MAP5 nº2012-06.
                   http://www.math-info.univ-paris5.fr/~moisan/papers/2012-06.pdf
  + [2] CUTASTRE : Rémy Abergel and Lionel Moisan, ``Accelerated A-contrario Detection of Smooth Trajectories'', Proceedings of EUSIPCO (2014).
                   http://www.eurasip.org/Proceedings/Eusipco/Eusipco2014/HTML/papers/1569926799.pdf

  Websites :

    + ASTRE : http://www.math-info.univ-paris5.fr/~moisan/astre/
    + CUTASTRE (under construction) : http://www.math-info.univ-paris5.fr/~moisan/cutastre/

  Versions :

  + v 1.0 : (12/2014) initial version from MegaWave implementation (Rémy Abergel)
  + v 1.1 : (01/2015) set 4th entry optional for the input sequence.
  + v 1.2 : (01/2015) debug precomputation of discrete acceleration & reduce RADIUS_MAX to 300
                      which ensures that the relative error between discrete and continous
                      acceleration is less than 1e-3. (Rémy Abergel)
  + v 1.3 : (03/2015) minor changes of indentation
                      modify displayed messages (see printf() calls)
                      remove define IDF, IDX, IDY, IDT. (Rémy Abergel)
  + v 1.4 : (06/2015) add website adresses. (Rémy Abergel)

  + v 1.5 : (08/2015, Rémy Abergel)

    + minor changes for the standard output (printf) at
      program execution.
    + update display usage and change some option symbols
      (option -X changed into -W (Width), option -Y changed
      into option -H (Height), option -e changed into -m
      (minscore)).
    + changed input format (comments are now allowed).
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define VERSION "1.5"

/* basic macros */
#define MIN(a,b)        (((a)>(b)) ? (b) : (a))
#define MAX(a,b)        (((a)>(b)) ? (a) : (b))
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Fatal error: Not enough memory.\n"); free_all(); exit(EXIT_FAILURE); }

/***************************************************************************************************
  COMPUTE_LOG10NFA macro : computes log10(NFA_T[acc(T)]) where T is a trajectoy of length l,
    ending in frame k (frames are numeroted from 0 to K-1), acc = acc(T) is the acceleration of
    trajectory T.
****************************************************************************************************/
#define COMPUTE_LOG10NFA(acc,k,l,chunksize)				\
  (log10Nchunks + log10tab[(chunksize)-1] + log10tab[(chunksize)-(l)] + \
   (cumsumlog10nppf[k]-((((k)-(l)+1)==0)?0:cumsumlog10nppf[(k)-(l)])) + \
   ((l)-2)*log10(acc) )

/***************************************************************************************************
  CALL_G macro : (see conventions below) allows to read or write G value at specified (k0,x0,y0,l0)
****************************************************************************************************/
#define CALL_G(k0,x0,y0,l0) G[k0][(y0)*Nptmax+(x0)][l0]

/***************************************************************************************************
  CALL_DIST : (see conventions below) allows to read or write Dist value at specified (k0,x0,y0)
****************************************************************************************************/
#define CALL_DIST(k0,x0,y0) dist[k0][(y0)*Nptmax+(x0)]

#ifndef M_PI
#define M_PI 3.14159265358979323
#endif

/* max radius (in pixels) for discrete acceleration computation (see note below) */
#define RADIUS_MAX 300

/* prototypes */
static void __init__(char *filenamein);
static void free_all(void);
static void free_G(void);
static void free_dist(void);
static void compute_dist(int);
static int compar(const void*, const void*);
static double compute_acc(int,int,int);
static double minimal_log10nfa(int,int*);
static void compute_G(int);
static void backtrack(int,int,double);

/* global variables */
static unsigned long int *dacc = NULL,d2max;
static int numTh,K,size,*ptseq = NULL,*xtab = NULL,*ytab = NULL,*nppf=NULL,*id0=NULL,*state=NULL,*toadd=NULL,*xm=NULL,*ym=NULL,*km=NULL,*lm=NULL,ntraj,ntraj0,lmin,nmax=10,nx,ny,C,O,Delta,Nptmax,Ki,Kext,Nchunks,*x1tox0=NULL,*length=NULL,*newid=NULL;
static double ***G = NULL,Omega,_INF=1./0.,log10Nchunks,*log10tab = NULL,*cumsumlog10nppf = NULL,maxspeed,maxspeed2,*g0 = NULL;
static char **dist = NULL,recompGneeded,maxspeedmode,vflag,lastchunk,*isx0 = NULL,*isx1 = NULL;

/***********************************************************************************************
 Notes about G computation
 -------------------------

   G[k0][y0*Nptmax+x0][l0] = G(x^k,y^{k-1},l) = smallest acceleration of a trajectory with
                                                length l, ending with link y^{k-1}->x^k within
                                                current chunk (chunk i).

   frames are numeroted from 0 to K-1 within the whole sequence (index k)
   chunks are numeroted from 0 to Nchunks-1
   C = size of a chunk (specified by the user)
   O = size of the overlap areas (specified by the user)
   Delta = C-O
   C{i} =  size of chunk i, C{i} = C for all chunks except eventually for the last chunk,
           as eventually C{Nchunks-1} <= C.
   first frame of chunk i has index k=i*Delta
   last  frame of chunk i has index k=i*Delta+C{i}-1
   n{k} = number of points in frame k
   Nptmax = max_k n{k}
   id0{k} = sum_{0<=j<k} n{j}

   k       = i*Delta+1 .. i*Delta+(C{i}-1)   Index of the frame that contains point x^k.
   x^k     = id0{k} .. id0{k}+n{k}-1         Index of point x^k in tabs (xtab,ytab).
   y^{k-1} = id0{k-1} .. id0{k-1}+n{k-1}-1   Index of point y^{k-1} in tabs (xtab,ytab).
   l       = 2 .. k-i*Delta+1                Length of the current trajectory.

   k0 = k-(i*Delta+1)     : 0 .. C{i}-2
   x0 = x^k-id0{k}        : 0 .. n{k}-1
   y0 = y^{k-1}-id0{k-1}  : 0 .. n{k-1}-1
   l0 = l-2               : 0 .. k-i*Delta-1

 Notes about dist computation
 ----------------------------

   dist[k0][y0*Nptmax+x0] = dist(x^k,y^{k-1}) = 1 if distance ||x^k - y^{k-1}|| <= maxspeed
                                              = 0 otherwise.

   dist is reinitialized at each change of chunk.
   i = index of current chunk.

   k       = i*Delta+1 .. i*Delta+(C{i}-1)   Index of the frame that contains point x^k.
   x^k     = id0{k} .. id0{k}+n{k}-1         Index of point x^k in tabs (xtab,ytab).
   y^{k-1} = id0{k-1} .. id0{k-1}+n{k-1}-1   Index of point y^{k-1} in tabs (xtab,ytab).

   k0 = k-(i*Delta+1)     : 0 .. C{i}-2
   x0 = x^k-id0{k}        : 0 .. n{k}-1
   y0 = y^{k-1}-id0{k-1}  : 0 .. n{k-1}-1


 Notes about discrete acceleration
 ---------------------------------

   As described in ref. [1], when we compute NFAs, we need to compute the areas of the
   discrete balls containing some pixel (rather than approximating this area by the
   area of the euclidian ball containing that pixel, which is problematic for instance
   if the pixel is the one at the origin, resulting in a null area).

   The discrete balls are the intersections of the Z^2 grid with the euclidian balls.

   We set a strictly positive number RADIUS_MAX that will give a bound for the computed
   areas. If we are given a radius r greater than RADIUS_MAX, we approximate the
   discrete ball area by M_PI*r^2 that is the euclidian ball area.

   The computation of the discrete acceleration is done within __init__ function, it
   results in a 1-D tab named dacc, such as dacc[x*x+y*y] equals the area of the
   discrete ball of radius r = sqrt(x*x+y*y) if r is less than RADIUS_MAX.

   Notice: we can usually decrease the value of RADIUS_MAX according to the size of
   the domain Omega or/and the use of an upper bound for the speed (option -S).

*************************************************************************************************/

/***********************************************************************************************/
/*             secondary modules (not directely dedicated to CUTASTRE computation)             */
/***********************************************************************************************/

/* usage displayer */
static void display_usage()
{
  printf("\nUsage: cutastre_noholes [--help] [--version] [-W width] [-H height] [-S maxspeed] [-c chunksize] [-o overlapsize] [-m minscore] [-T numthreads] [-v] filenamein filenameout\n\n");
  printf("   --help         : display help\n");
  printf("   --version      : display program version\n");
  printf("   -W width       : set the width (in pixels) of the image domain\n                    (default = Xmax-Xmin+1, X(min/max) = smallest/highest \n                    X-coordinate (horizontal axis) of the input sequence)\n");
  printf("   -H height      : set the height (in pixels) of the image domain\n                    (default = Ymax-Ymin+1, Y(min/max) = smallest/highest\n                    Y-coordinate (vertical axis) of the input sequence)\n");
  printf("   -S maxspeed    : only detect trajectories with maximal speed (that is\n                    the maximal distance (in pixels) between two consecutive\n                    points) less than maxpseed (may drastically reduce the\n                    execution time)\n");
  printf("   -c chunksize   : split the frames of the sequence into chunks of specified\n                    size (in number of frames), must be >=3 (default = 20)\n");
  printf("   -o overlapsize : set the overlap size (in number of frames) between two \n                    consecutive chunks, must be >= 2 and < c (default =\n                    maximal value between 2 and floor(c/2))\n");
  printf("   -m minscore    : (default = 0) set the score threshold, only trajectories\n                    with a score higher than minscore will be extracted\n");
  printf("   -T numthreads  : (default = 1) set the number of OpenMP threads\n");
  printf("   -v             : verbose mode\n");
  printf("   filenamein     : path to the input point set sequence file\n");
  printf("   filenameout    : path to the output point set sequence file\n\n");
}


/* main initialization module (retrieve input data, memory allocation, etc.) */
static void __init__(char *filenamein)
{
  int k0,k,n,id,xmin,ymin,xmax,ymax,*tmp,x,y,x0,Nptmax2,c,L,e,N,len,lenMax,nlines,ncomments;
  char *line;
  unsigned long int *d2map;
  double accmax;
  FILE *file;

  size = nlines = ncomments = len = lenMax = 0;
  xmin = ymin =  _INF; xmax = ymax = -_INF; lmin = 3;

  /* open input file, count number of (nonempty) lines and compute the maximal length of a line */
  if(NULL == (file = fopen(filenamein,"r"))) {
    printf("Fatal error: could not open file '%s'.\n",filenamein);
    free_all(); exit(EXIT_FAILURE);
  }
  while((c = fgetc(file)) != EOF) {
    len++;
    if(c == '\n') { nlines++; lenMax = MAX(len,lenMax); len=0; }
  }
  rewind(file);
  lenMax++;
  if(NULL == (line = (char*) malloc (lenMax*sizeof(char)))) {
    printf("Fatal error: Not enough memory.\n");
    fclose(file); free_all(); exit(EXIT_FAILURE);
  }

  /* count the number of lines starting with character '#' (comments) */
  for (n=0;n<nlines;n++) {
    if(NULL == fgets(line,lenMax,file)) {
      printf("An error occured while reading input file.\n");
      fclose(file); free(line); free_all(); exit(EXIT_FAILURE);
    }
    if(line[0]=='#') ncomments++;
  }
  rewind(file);

  /* allocate memory for the point sequence and retrieve data */
  size = nlines - ncomments;
  if(NULL == (ptseq = (int*) malloc (3*size*sizeof(int)))) {
    printf("Fatal error: Not enough memory.\n");
    fclose(file); free(line); free_all(); exit(EXIT_FAILURE);
  }

  /* retrieve data as it was a Flist (MegaWave type) */
  for(n=c=0;c<nlines;c++) {
    if(NULL == fgets(line,lenMax,file)) {
      printf("An error occured while reading input file.\n");
      fclose(file); free(line); free_all(); exit(EXIT_FAILURE);
    }
    if(line[0] != '#') {
      if (n < size) e = sscanf(line,"%d %d %d",&ptseq[n*3],&ptseq[n*3+1],&ptseq[n*3+2]);
      if ((e < 3)||(n>=size)) {
	printf("Fatal error: could not correctly retrieve data from the input point sequence.\n");
	printf("Please check the format of your input point set sequence, each line must\nstart with three integers ``<frameid> <x> <y>'' (separated with space) OR\nstart with character '#' (comment). Notice that empty lines are not allowed!\n");
	fclose(file); free(line); free_all(); exit(EXIT_FAILURE);
      }
      n++;
    }
  }
  fclose(file); free(line); line = NULL;

  /**************************/
  /*   build ASTRE inputs   */
  /**************************/

  /* retrieve number of frames */
  for(K=n=0;n<size;n++) K = MAX(K,ptseq[n*3]);
  K++; if(K<3) { printf("error: input sequence must contain at list 3 frames\n"); free_all(); exit(EXIT_FAILURE); }

  /* compute Nchunks (number of chunks) */
  Delta = C-O;
  Nchunks=1+(int)floor((double)(K-1)/(double)Delta);
  if (Nchunks > 1) while ((Nchunks-1)*Delta+O-1 >= K-1) Nchunks --;
  log10Nchunks = log10(Nchunks);

  /* retrieve the number of point per frame */
  ASSERT_ALLOC(nppf = (int*) calloc (K,sizeof(int)));
  for(n=0;n<size;n++) nppf[ptseq[n*3]]++;

  /* compute Nptmax = max_k nppf[k] */
  for(Nptmax=0,k=0;k<K;k++) Nptmax = MAX(Nptmax,nppf[k]);
  Nptmax2 = Nptmax * Nptmax;

  /* compute id0[k] for k = 0..K-1 (first index corresponding to frame k in xtab, ytab) */
  ASSERT_ALLOC(id0 = (int*) malloc (K*sizeof(int)));
  for(id0[0]=0,k=1;k<K;k++) id0[k] = id0[k-1] + nppf[k-1];

  /* initialize state */
  ASSERT_ALLOC(state = (int*) malloc(size*sizeof(int)));
  for(n=0;n<size;n++) state[n] = -1;

  /* fill xtab & ytab */
  ASSERT_ALLOC(xtab = (int*) malloc(size*sizeof(int)));
  ASSERT_ALLOC(ytab = (int*) malloc(size*sizeof(int)));
  ASSERT_ALLOC(tmp  = (int*) calloc(K,sizeof(int)));
  for(n=0;n<size;n++) {
    k = ptseq[n*3];
    id = id0[k]+tmp[k];
    xtab[id] = ptseq[n*3+1];
    ytab[id] = ptseq[n*3+2];
    xmin = MIN(xmin,xtab[n]);
    ymin = MIN(ymin,ytab[n]);
    xmax = MAX(xmax,xtab[n]);
    ymax = MAX(ymax,ytab[n]);
    tmp[k]++;
  }
  /* close file & free memory (not used anymore) */
  free(tmp); tmp = NULL; free (ptseq); ptseq = NULL;

  /* retrieve or control domain dimensions */
  if(nx == -1) nx = xmax-xmin+1;
  else if (nx < xmax-xmin+1) {
    printf("error: unconsistent width for the image domain (points of the input sequence must be included in the domain).\n");
    free_all(); exit(EXIT_FAILURE);
  }
  if(ny == -1) ny = ymax-ymin+1;
  else if (ny < ymax-ymin+1) {
    printf("error: unconsistent height for the image domain (points of the input sequence must be included in the domain).\n");
    free_all(); exit(EXIT_FAILURE);
  }
  Omega = (double) (nx*ny);

  /* allocate memory (toadd, xm, ym, x0m, y0m, km, lm, ...) */
  ASSERT_ALLOC(toadd  = (int*)    malloc (MIN(K,C)*sizeof(int)));
  ASSERT_ALLOC(xm     = (int*)    malloc (nmax*sizeof(int)));
  ASSERT_ALLOC(ym     = (int*)    malloc (nmax*sizeof(int)));
  ASSERT_ALLOC(km     = (int*)    malloc (nmax*sizeof(int)));
  ASSERT_ALLOC(lm     = (int*)    malloc (nmax*sizeof(int)));
  ASSERT_ALLOC(newid  = (int*)    malloc (size*sizeof(int)));
  ASSERT_ALLOC(length = (int*)    malloc (size*sizeof(int)));    // NB : we can save memory by using a sharper size (size = 2 * max_k nppf[k] is enough)
  ASSERT_ALLOC(x1tox0 = (int*)    malloc (size*sizeof(int)));    // NB : we can save memory by using a sharper size (size = 2 * max_k nppf[k] is enough)
  ASSERT_ALLOC(isx0   = (char*)   calloc (size,sizeof(char)));   // NB : we can save memory by using a sharper size (size = 2 * max_k nppf[k] is enough)
  ASSERT_ALLOC(isx1   = (char*)   calloc (size,sizeof(char)));   // NB : we can save memory by using a sharper size (size = 2 * max_k nppf[k] is enough)
  ASSERT_ALLOC(g0     = (double*) malloc (size*sizeof(double))); // NB : we can save memory by using a sharper size (size = 2 * max_k nppf[k] is enough)

  /***************************/
  /* memory allocation for G */
  /***************************/
  ASSERT_ALLOC(G = (double***) malloc ((MIN(K,C)-1)*sizeof(double**)));
  for(k0=0; k0<(MIN(K,C)-1); k0++) {
    ASSERT_ALLOC(G[k0] = (double**) malloc (Nptmax2*sizeof(double*)));
    for(x0=0; x0<Nptmax2; x0++) ASSERT_ALLOC(G[k0][x0] = (double*) malloc (MIN(K,C)*sizeof(double)));
  }

  /******************************/
  /* memory allocation for dist */
  /******************************/
  ASSERT_ALLOC(dist = (char**) malloc ((MIN(K,C)-1)*sizeof(char*)));
  for(k0=0; k0<(MIN(K,C)-1); k0++) ASSERT_ALLOC(dist[k0] = (char*) malloc (Nptmax2*sizeof(char)));

  /*********************************************/
  /* tabulate log10 and cumsumlog10nppf values */
  /*********************************************/
  ASSERT_ALLOC(log10tab = (double*) malloc (((Nchunks==1)?K:(2*C-O))*sizeof(double))); // log10tab[n]  = log10(n+1)
  ASSERT_ALLOC(cumsumlog10nppf = (double*) malloc (K*sizeof(double))); // cumsum_log10nppf[n] = sum_{i=0}^n log10(nppf[n])
  for(n=0;n<((Nchunks==1)?K:(2*C-O));n++) log10tab[n] = log10(n+1);
  for(cumsumlog10nppf[0]=log10(nppf[0]),n=1;n<K;n++) cumsumlog10nppf[n] = log10(nppf[n]) + cumsumlog10nppf[n-1];

  /*************************************/
  /*   compute discrete acceleration   */
  /*************************************/
  if(vflag) printf("\nInitialization: computing discrete acceleration...\n");
  accmax = MIN((double) RADIUS_MAX,2*sqrt(nx*nx+ny*ny)); if(maxspeedmode) accmax = MIN(accmax,maxspeed*2.);
  L = (unsigned long int) ceil(accmax); N = L*L;
  ASSERT_ALLOC(d2map = (unsigned long int*) malloc (N*sizeof(unsigned long int)));
  for(y=1;y<=L;y++) for(x=1;x<=L;x++) d2map[(y-1)*L+(x-1)] = (unsigned long int) (x*x+y*y);
  qsort((void *)d2map,N,sizeof(unsigned long int),compar);
  d2max = (unsigned long int) floor(accmax*accmax);
  ASSERT_ALLOC(dacc = (unsigned long int*) calloc ((1+d2max),sizeof(unsigned long int)));
  /* set dacc[d2] = Cardinal of the set {(x,y), x >= 1, y >= 1, x^2+y^2 = d2} */
  for(dacc[0]=0.,n=0;n<N;n++) if(d2map[n] <= d2max) dacc[d2map[n]]++;
  for(n=1;n<=d2max;n++) dacc[n] += dacc[n-1]; // cumsum dacc, now dacc[d2] = Cardinal of the set {(x,y), x >= 1, y >= 1, x^2+y^2 <= d2}
  /* extend dacc to the whole ball (up to here, only the first cadran has been considered) */
  for(n=0;n<=d2max;n++) dacc[n] = 4*(dacc[n]+(unsigned long int) floor(sqrt(n)))+1; // point (x=y=0) is added 1 time, the others are added 4 times.
  free(d2map); d2map = NULL;
  if(vflag) printf("...computation done.\n");
}

/* free all memory (must be called before exit) */
static void free_all()
{
  if(NULL != dacc) free(dacc);
  if(NULL != ptseq) free(ptseq);
  if(NULL != xtab) free(xtab);
  if(NULL != ytab) free(ytab);
  if(NULL != nppf) free(nppf);
  if(NULL != id0) free(id0);
  if(NULL != state) free(state);
  if(NULL != toadd) free(toadd);
  if(NULL != xm) free(xm);
  if(NULL != ym) free(ym);
  if(NULL != km) free(km);
  if(NULL != lm) free(lm);
  if(NULL != x1tox0) free(x1tox0);
  if(NULL != length) free(length);
  if(NULL != newid) free(newid);
  if(NULL != log10tab) free(log10tab);
  if(NULL != cumsumlog10nppf) free(cumsumlog10nppf);
  if(NULL != isx0) free(isx0);
  if(NULL != isx1) free(isx1);
  if(NULL != g0) free(g0);
  if(NULL != G) free_G();
  if(NULL != dist) free_dist();
}

/* free memory allocated for G */
static void free_G()
{
  int k0,x0,Nptmax2=Nptmax*Nptmax;
  for(k0=0; k0<(MIN(K,C)-1); k0++) {
    for(x0=0; x0<Nptmax2; x0++) free(G[k0][x0]);
    free(G[k0]);
  }
  free(G); G = NULL;
}

/* brief description: free memory allocated for dist */
static void free_dist()
{
  int k0,x0;
  for(k0=0; k0<(MIN(K,C)-1); k0++) free(dist[k0]);
  free(dist); dist = NULL;
}

/* controle distance between points (consecutives frames) within chunk i */
static void compute_dist(int i)
{
  int k,k0,x,x0,y,y0,Nptmax2=Nptmax*Nptmax;
  double dx,dy;

  if(maxspeedmode)
    for(k=i*Delta+1,k0=0; k<MIN(K,i*Delta+C); k0++,k++)
      for(x=id0[k],x0=0; x0<nppf[k]; x0++,x++)
	for(y=id0[k-1],y0=0; y0<nppf[k-1]; y0++,y++) {
	  dx = ((double) xtab[x]) - ((double) xtab[y]); dy = ((double) ytab[x]) - ((double) ytab[y]);
	  CALL_DIST(k0,x0,y0) = (dx*dx+dy*dy <= maxspeed2);
	}
  else { for(k0=0;k0<(MIN(K,C)-1);k0++) for(x0=0;x0<Nptmax2;x0++) dist[k0][x0] = 1; }
}

/* comparison function for qsort() */
static int compar(const void *a, const void *b)
{
  int v = *((unsigned long int const *)a)-*((unsigned long int const *)b);
  return (v>0 ? 1 : ((v<0) ? -1 : 0));
}

/*****************************************************************************************/
/*                  primary modules (dedicated to CUTASTRE computation)                  */
/*****************************************************************************************/

/*************************************************************************************
   compute_acc : compute discrete acceleration of a 3-tuple (x,y,z) of point such as
     x = (xtab[idx],ytab[idx]), y = (xtab[idy],ytab[idy]), z = (xtab[idz],ytab[idz])
**************************************************************************************/
static double compute_acc(int idx, int idy, int idz)
{
  unsigned long int d2,x,y;
  x = (unsigned long int) (xtab[idx] - 2*xtab[idy] + xtab[idz]);
  y = (unsigned long int) (ytab[idx] - 2*ytab[idy] + ytab[idz]);
  d2 = x*x+y*y;
  return (d2<=d2max) ? dacc[d2]/Omega : M_PI*((double) d2)/Omega;
}

/* Dynamic programming computation of G within chunk i */
static void compute_G(int i)
{
  int k0,k,x0,x,y0,y,z0,z,l0,l;
  double a,a1,a2;
  char izx0,izx1;
  for(k0=0,k=i*Delta+1; k<MIN(K,i*Delta+C); k++,k0++) {
#pragma omp parallel for num_threads(numTh) shared(nppf,k,k0,id0,state,G,Omega,x1tox0,Delta,O) private(x,x0,y,y0,l,l0,z,z0,a,a1,a2,izx0,izx1)
    for(x0=0; x0<nppf[k]; x0++) {
      x = x0 + id0[k];
      izx1 = isx1[x] && (k==(i+1)*Delta+O-1);
      izx0 = isx0[x] && (k==(i+1)*Delta+O-2);
      if((state[x] == -1) || izx0 || izx1) {
	if(izx1) { // compute G(x,y,l) only for y=x1tox0[x] (links y->x cannot exist when y is different from x1tox0[x])
	  y = x1tox0[x]; y0 = y-id0[k-1]; CALL_G(k0,x0,y0,0) = 1./Omega;
	  for(l=3,l0=1; l0 < k-i*Delta; l++,l0++) {
	    CALL_G(k0,x0,y0,l0) = _INF;
	    for(z0=0,z=id0[k-2]; z0<nppf[k-2]; z++,z0++)
	      if((state[z] == -1) && CALL_DIST(k0-1,y0,z0)) {
		a1=compute_acc(x,y,z); a2=CALL_G(k0-1,y0,z0,l0-1); a=MAX(a1,a2); a2=CALL_G(k0,x0,y0,l0); CALL_G(k0,x0,y0,l0)=MIN(a,a2);
	      }
	  }
	}
	else // x is not a x1-point, compute G(x,y,l) for all (y,l) except when y is a x0 or x1-point
	  for(y0=0,y=id0[k-1]; y0<nppf[k-1]; y++,y0++)
	    if((state[y] == -1) && CALL_DIST(k0,x0,y0)) { // y is not a x0-point nor a x1-point because state[y]==-1
	      CALL_G(k0,x0,y0,0) = 1./Omega;
	      for(l=3,l0=1; l0 < k-i*Delta; l++,l0++) {
		CALL_G(k0,x0,y0,l0) = _INF;
		for(z0=0,z=id0[k-2]; z0<nppf[k-2]; z++,z0++)
		  if((state[z] == -1) && CALL_DIST(k0-1,y0,z0)) {
		    a1=compute_acc(x,y,z); a2=CALL_G(k0-1,y0,z0,l0-1); a=MAX(a1,a2); a2=CALL_G(k0,x0,y0,l0); CALL_G(k0,x0,y0,l0)=MIN(a,a2);
		  }
	      }
	    }
      }
    }
  }
}

/* Find the minimal log10(NFA) value among all trajectories within chunk i, and count the number (n) of trajectories having optimal NFA */
static double minimal_log10nfa(int i, int *n)
{
  int k0,k,x,x0,y,y0,l,l0,z,z0,l1;
  double m=_INF,log10nfa,a0,a1;
  char izx1;
  *n=1;
  // find minimal NFA //
  for(k=(i==0) ? 2 : i*Delta+O, k0=k-i*Delta-1; k<MIN(K,i*Delta+C); k++,k0++)
    for(x0=0,x=id0[k]; x0<nppf[k]; x++,x0++) {
      izx1 = (k==(i+1)*Delta+O-1) && isx1[x];
      if ((state[x]==-1) || izx1) {
	if(izx1) { // only consider link y->x fox y=x1tox0[x], and take account of trajectory extension
	  y = x1tox0[x]; y0 = y-id0[k-1]; a0 = g0[y]; l1 = length[y];
	  for(l=lmin,l0=lmin-2; l0<(k-i*Delta); l++,l0++) {
	    a1 = CALL_G(k0,x0,y0,l0); log10nfa = COMPUTE_LOG10NFA(MAX(a0,a1),k+l1-2,l+l1-2,Kext);
	    if ((m==log10nfa)&&(*n<nmax)) { xm[*n]=x; ym[*n]=y; km[*n]=k; lm[*n]=l; *n=*n+1; }
	    else if (m>log10nfa) { *n=1; xm[0]=x; ym[0]=y; km[0]=k; lm[0]=l; m=log10nfa; }
	  }
	}
	else { // x is not a x1 nor a x0-point (because state[x]==-1), only consider links y->x when y is not a x0 nor a x1-point, no trajectory extension here
	  for(y0=0,y=id0[k-1]; y0<nppf[k-1]; y0++,y++)
	    if((state[y] == -1) && CALL_DIST(k0,x0,y0)) // y is not a x0-point nor a x1-point because state[y]==-1
	      for(l=lmin,l0=lmin-2; l0<(k-i*Delta); l++,l0++) {
		log10nfa = COMPUTE_LOG10NFA(CALL_G(k0,x0,y0,l0),k,l,Ki);
		if ((m==log10nfa)&&(*n<nmax)) { xm[*n]=x; ym[*n]=y; km[*n]=k; lm[*n]=l; *n=*n+1; }
		else if (m>log10nfa) { *n=1; xm[0]=x; ym[0]=y; km[0]=k; lm[0]=l; m=log10nfa; }
	      }
	}
      }
    }
  // return minimal log10(nfa) //
  return m;
}

/**********************************************************************************************
  backtrack(i,n,log10nfa) :

    backtrack optimal trajectory within chunk i, with length l, log10(nfa) = log10nfa,
    ending with link y^{k-1}->x^k, such as  k = km[n], x^k = xm[n], y^{k-1} = ym[n], l = lm[n].

***********************************************************************************************/
static void backtrack(int i, int n, double log10nfa)
{
  char extend = (state[xm[n]] >= 0), foundx0 = 0;
  int x=xm[n],y=ym[n],z,k=km[n],l=lm[n],y0=y-id0[k-1],z0,k0=k-(i*Delta+1),l0=l-2,id=(extend)?state[x]:ntraj,lth=2,lth0,zopt,z0opt,tobex0,tobex1;
  double mm,a,a1,a2,am=0,amax;

  // if head has not been detected, change state of ending points, otherwise return //
  recompGneeded = ((state[x] == -2) || (state[y] == -2)); if(recompGneeded) return;
  toadd[l-1] = x; toadd[l-2] = y;

  // backtrack predecessors //
  while (l>2) {
    mm = _INF;
    for(z0=0,z=id0[k-2]; z0<nppf[k-2]; z++,z0++)
      if((state[z]<0) && CALL_DIST(k0-1,y0,z0)) {
	a1 = compute_acc(x,y,z); a2 = CALL_G(k0-1,y0,z0,l0-1); a = MAX(a1,a2);
	if(a < mm) { mm=a; zopt = z, z0opt = z0; }
      }
    recompGneeded = (state[zopt] == -2); if(recompGneeded) return;
    toadd[l-3] = zopt; lth++; am = MAX(am,mm);
    if( (!foundx0) && (k-2==i*Delta+O-2) ) { foundx0=1; tobex0=zopt; tobex1=y; amax=am; lth0=lth; }
    x = y; y = zopt; y0 = z0opt; l--; l0--; k--; k0--;
  }

  // if recompGneeded is still 0, the tracked trajectory is valid //
  for(l=0;l<lm[n];l++) { state[toadd[l]] = -2; newid[toadd[l]] = id; }
  if(foundx0) { isx0[tobex0]=isx1[tobex1]=1; x1tox0[tobex1]=tobex0; g0[tobex0]=amax; length[tobex0]=lth0; }

  // in case of an extension, we must change x0-state or x1-state of the head points //
  if(extend && isx1[xm[n]]) isx1[xm[n]] = isx0[ym[n]] = 0;

  /* update ntraj (number of detected trajectories) */
  if (vflag && extend)  printf("  + Trajectory %d has been extended: score = %.10g\n",id,-log10nfa);
  if (vflag && !extend) printf("  + Trajectory %d has been extracted: score = %.10g\n",id,-log10nfa);
  if (!extend) ntraj++;
}


/************ MAIN MODULE ************/
int main(int argc, char **argv)
{
  int i,n,k,err;
  char *Wvalue = NULL, *Hvalue = NULL, *Tvalue = NULL, *cvalue = NULL, *ovalue = NULL, *mvalue = NULL, *Svalue = NULL, *filenamein = NULL, *filenameout = NULL;
  double minscore,maxlog10nfa,m;
  FILE *file;

  vflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-v") == 0) vflag = 1;
    else if ((strcmp(argv[i],"--help") == 0) || (strcmp(argv[i],"-?") == 0)) {
      display_usage();
      /*
      printf("Input/output files :\n\n   They are simple text files where each line describes a point of the\n   sequence. All lines must contain 3 or 4 integers named <frameid> <x> <y>\n   <trajid>.\n\n");
      printf("   + <frameid> : identifier of the frame containing the point (frames are\n                 numeroted from 0)\n");
      printf("   + <x>       : X-coordinate of the point\n");
      printf("   + <y>       : Y-coordinate of the point\n");
      printf("   + <trajid>  : [optional entry for the input file] identifier of the\n                 trajectory to which the point belongs (never used by the\n                 algorithm), the detected trajectories are numeroted from 0,\n                 while the spurious points (noise) have trajid = -1.\n\n");
      */
      printf("Input files : TODO\n\n");
      printf("Output files : TODO\n\n");
      printf("   Example (first lines of the snow sequence): \n\n   0 292 204 4\n   0 247 304 51\n   0 358 295 74\n       ...\n\n");
      printf("Main references : \n\n");
      printf("   + ASTRE    : Maël Primet and Lionel Moisan, ``Point tracking: an a-contrario\n                approach'',  preprint MAP5 nº2012-06.\n                http://www.math-info.univ-paris5.fr/~moisan/papers/2012-06.pdf\n\n");
      printf("   + CUTASTRE : Rémy Abergel and Lionel Moisan, ``Accelerated A-contrario\n                Detection of Smooth Trajectories'', Proceedings of EUSIPCO\n                (2014).\n                http://www.eurasip.org/Proceedings/Eusipco/Eusipco2014/HTML/papers/1569926799.pdf\n\n");
      return EXIT_SUCCESS;
    }
    else if (strcmp(argv[i],"--version") == 0) {
      printf("\n*****************************************************************************\nCUTASTRE Accelerated A-contrario Detection of Smooth TRajectories version %s\n*****************************************************************************\n\nCopyright (C) 2014 Rémy Abergel (remy.abergel AT gmail.com) and Lionel Moisan\n(Lionel.Moisan AT parisdescartes.fr) Université Paris Descartes, Laboratoire\nMAP5, CNRS UMR 8145.\n\nThis program is free software: you can redistribute it and/or modify it under\nthe terms of the GNU General Public License as published by the Free Software\nFoundation, either version 3 of the License, or (at your option) any later\nversion.\n\nThis program is distributed in the hope that it will be useful, but WITHOUT\nANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\nFOR A PARTICULAR PURPOSE. See the GNU General Public License for more\ndetails.\n\nYou should have received a copy of the GNU General Public License along with\nthis program. If not, see <http://www.gnu.org/licenses/>.\n\n",VERSION);
      return EXIT_SUCCESS;
    }
    else if (strcmp(argv[i],"-W") == 0) {
      if(i==argc-1) { printf("\nOption -W requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	Wvalue = argv[i+1]; err = sscanf(Wvalue,"%d",&nx); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -W\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-H") == 0) {
      if(i==argc-1) { printf("\nOption -H requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	Hvalue = argv[i+1]; err = sscanf(Hvalue,"%d",&ny); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -H\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-T") == 0) {
      if(i==argc-1) { printf("\nOption -T requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	Tvalue = argv[i+1]; err = sscanf(Tvalue,"%d",&numTh); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -T\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-c") == 0) {
      if(i==argc-1) { printf("\nOption -c requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	cvalue = argv[i+1]; err = sscanf(cvalue,"%d",&C); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -c\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-o") == 0) {
      if(i==argc-1) { printf("\nOption -o requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	ovalue = argv[i+1]; err = sscanf(ovalue,"%d",&O); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -o\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-m") == 0) {
      if(i==argc-1) { printf("\nOption -m requires an argument\n"); display_usage(); return EXIT_FAILURE; }
      else {
	mvalue = argv[i+1]; err = sscanf(mvalue,"%lf",&minscore); i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -m\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-S") == 0) {
      if(i==argc-1) { printf("\nOption -S requires an argument\n"); return EXIT_FAILURE; }
      else {
	Svalue = argv[i+1]; err = sscanf(Svalue,"%lf",&maxspeed); maxspeed2 = maxspeed*maxspeed; i++;
	if(err != 1) { printf("\nCould not retrieve properly the argument of option -S\n"); display_usage(); return EXIT_FAILURE; }
      }
    }
    else if (argv[i][0] == '-') { printf("\nUnrecognized option `%s'.\n",argv[i]); display_usage(); return EXIT_FAILURE; }
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { printf("\nCould not retrieve properly input arguments.\n"); display_usage(); return EXIT_FAILURE; }
  }

  /* set default values */
  if(NULL == Wvalue) nx = -1.; // default value will be set inside __init__ function.
  if(NULL == Hvalue) ny = -1.; // default value will be set inside __init__ function.
  if(NULL == Tvalue) numTh = 1;
  if(NULL == cvalue) C = 20;
  if(NULL == ovalue) O = MAX(2,C/2);
  if(NULL == mvalue) minscore = 0;
  maxspeedmode = (NULL != Svalue);
  maxlog10nfa = -minscore;

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenameout)) {
    if (NULL == filenamein) printf("\nerror: filenamein missing\n");
    else if(NULL == filenameout) printf("\nerror: filenameout missing\n");
    display_usage(); free_all();
    return EXIT_FAILURE;
  }
  if (numTh <= 0) { printf("error: (unconsistent input numTh) you must select at least one thread.\n"); free_all(); return EXIT_FAILURE; }
  if (C < 3) { printf("error: (unconsistent input chunksize) chunk size must be greater or equal to 3.\n"); free_all(); return EXIT_FAILURE; }
  if (O < 2) { printf("error: (unconsistent input overlapsize) overlap size must be greater or equal to 2.\n"); free_all(); return EXIT_FAILURE; }
  if (O >= C) { printf("error: (unconsistent input overlapsize) overlap size must be less than chunksize.\n"); free_all(); return EXIT_FAILURE; }
  if (maxspeedmode && maxspeed <= 0) { printf("fatal error: (unconsistent input maxspeed) maxspeed must be positive.\n"); free_all(); return EXIT_FAILURE; }

  /* main initialization (retrieve data from the input sequence of points, memory allocation, etc.) */
  __init__(filenamein);
  if (vflag) printf("\nInitialization done:\n\n  + number of frames = %d\n  + dimensions of the image domain = %d x %d\n  + number of chunks = %d\n  + chunks-size = %d\n  + overlap-size = %d\n\n",K,nx,ny,Nchunks,C,O);

  /****************************************************/
  /*                CUTASTRE MAIN LOOP                */
  /****************************************************/
  for(ntraj=0,lastchunk=1,i=Nchunks-1;i>=0;i--) {
    compute_dist(i);
    Ki = (lastchunk)? (K-i*Delta) : C;
    Kext = (i==Nchunks-2)? (K-i*Delta) : 2*C-O;
    if(vflag) printf("* Trajectories are being processed within chunk %d (frames %d to %d)\n",i+1,i*Delta+1,MIN(i*Delta+C-1,K-1)+1);
    do {
      compute_G(i);
      m = minimal_log10nfa(i,&n);
      recompGneeded = 0; ntraj0 = ntraj;
      while ((!recompGneeded)&&(m<=maxlog10nfa)) {
	for(k=0;k<n;k++) backtrack(i,k,m);
	if (!recompGneeded) m = minimal_log10nfa(i,&n);
	else if (vflag) printf("  >>> 2 trajectories are sharing the same point: now stop backtrack an recompute G\n");
      }
      for(k=i*Delta;k<MIN(K,i*Delta+C);k++) for(n=id0[k];n<id0[k]+nppf[k];n++) if(state[n]==-2) state[n] = newid[n];
    } while (m <= maxlog10nfa);
    if (vflag) printf("* chunk %d: best score detected = %.10g < %f, stop detection !\n\n",i+1,-m,minscore);
    // if i>0 set state of tail-points (within the overlap area, except its two last frames) back to -1 //
    if(i>0) for(k=i*Delta;k<MIN(K,i*Delta+O-2);k++) for(n=id0[k];n<id0[k]+nppf[k];n++) state[n] = -1;
    lastchunk = 0;
  }
  if (vflag) printf("Number of detected trajectories = %d\n\n",ntraj);
  /**************************************************/
  /*                END OF MAIN LOOP                */
  /**************************************************/

  /* write output */
  if(NULL == (file = fopen(filenameout,"w"))) {
    printf("error: could not open file '%s'\n",filenameout);
    free_all(); return EXIT_FAILURE;
  }
  for(n=0,k=0;k<K;k++) for(i=id0[k];i<id0[k]+nppf[k];i++,n++) fprintf(file,"%d %d %d %d\n",k,xtab[i],ytab[i],state[i]);
  fclose(file);

  /* free memory */
  free_all();

  return EXIT_SUCCESS;
}
